// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Resource Utilization Point Implementation                        ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "rtppacket.h"
#include "resourceutilizationpoint.h"


namespace Coral {


// ###### Reset #############################################################
void ResourceUtilizationPoint::reset()
{
   Bandwidth   = 0;
   Utilization = 0.0;
   Cost        = 0.0;
   TaskID      = 0;
   FrameRate   = 0.0;
   Layers      = 0;
   for(cardinal i = 0;i < RTPConstants::RTPMaxQualityLayers;i++) {
      LayerBandwidthInfo[i].reset();
   }
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const ResourceUtilizationPoint& rup)
{
   char str[300 + (RTPConstants::RTPMaxQualityLayers * 64)];
   char lstr[64];

   // ====== Print header ===================================================
   const cardinal frameRate = (cardinal)rint(rup.FrameRate * 10.0);
   snprintf((char*)&str, 256,
            "Task #%02d: U=%1.3f C=%6f   FPS=%2d.%d BPS=%7d   (",
            rup.TaskID,
            rup.Utilization, rup.Cost,
            frameRate / 10, frameRate % 10,
            rup.Bandwidth);

   // ====== Print each layer's bandwidth ===================================
   for(cardinal i = 0;i < rup.Layers;i++) {
      snprintf((char*)&lstr, sizeof(lstr) - 2, "L%d=%7d",
               i,
               rup.LayerBandwidthInfo[i].BytesPerSecond);
      if(i < (rup.Layers - 1)) {
         strcat((char*)&lstr,", ");
      }
      strcat((char*)&str,(char*)&lstr);
   }

   os << str << ")";
   return(os);
}


// ###### Sort resource/utilization list ####################################
void ResourceUtilizationPoint::sortResourceUtilizationList(
                                  ResourceUtilizationPoint* rup,
                                  const integer            start,
                                  const integer            end)
{
   const double v = rup[(start + end) / 2].Utilization;
   integer i      = start;
   integer j      = end;

   do {
      while(rup[i].Utilization < v) i++;
      while(rup[j].Utilization > v) j--;
      if(i <= j) {
         swapResourceUtilizationPoints(rup[i],rup[j]);
         i++;
         j--;
      }
   } while(j >= i);

   if(start < j) {
      sortResourceUtilizationList(rup,start,j);
   }
   if(i < end) {
      sortResourceUtilizationList(rup,i,end);
   }
}


// ###### Compute convex hull using Graham Scan algorithm ###################
cardinal ResourceUtilizationPoint::grahamScanResourceUtilizationList(
                                      ResourceUtilizationPoint* rup,
                                      const cardinal            count)
{
   // ResourceUtilizationPoint:
   //    x = Resource (Bandwidth)
   //    y = Utilization

   cardinal i;
   cardinal min;
   cardinal M;

   for(min = 0, i = 1;i < count;i++) {
      if(rup[i].Utilization < rup[min].Utilization) {
         min = i;
      }
   }

   for(i = 0;i < count;i++) {
      if(rup[i].Utilization == rup[min].Utilization) {
         if(rup[i].Bandwidth > rup[min].Bandwidth) {
            min = i;
         }
      }
   }
   swapResourceUtilizationPoints(rup[0],rup[min]);
   sortResourceUtilizationList(rup,0,count - 1);

   for(M = 2, i = 3; i < count;i++) {
      while(ccw(rup[M - 1],rup[M],rup[i]) >= 0) {
              // ????????????????? M, M - 1, i    >= 0 ???????????????
         M--;
         if(M == 0) {
            break;
         }
      }
      M++;
      swapResourceUtilizationPoints(rup[i],rup[M]);
   }
   return(M + 1);
}


// ###### Merge resource/utilization lists ##################################
cardinal ResourceUtilizationPoint::mergeResourceUtilizationLists(
                                       ResourceUtilizationPoint*  destination,
                                       ResourceUtilizationPoint** listArray,
                                       const cardinal*            listSizeArray,
                                       const cardinal             listCount)
{
   cardinal pos = 0;
   for(cardinal i = 0;i < listCount;i++) {
      for(cardinal j = 0;j < listSizeArray[i];j++) {
         destination[pos] = listArray[i][j];
         pos++;
      }
   }

   sortResourceUtilizationList(destination,0,pos - 1);
   return(pos);
}


}
