// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Frame Rate Scalability Implementation                      ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "traceframeratescalability.h"
#include "utilityfunction.h"


namespace Coral {


// ###### Constructor #######################################################
TraceFrameRateScalability::TraceFrameRateScalability()
{
   TraceReader = NULL;
}


// ###### Destructor ########################################################
TraceFrameRateScalability::~TraceFrameRateScalability()
{
}


// ###### Initialize ########################################################
void TraceFrameRateScalability::initFrameRateScalability(
                                     TDTFReader*  traceReader,
                                     const card64 position)
{
   TraceReader = traceReader;
   Position    = position;
}


// ###### Get frame rate scalability class name #############################
const char* TraceFrameRateScalability::getFrameRateScalabilityClass() const
{
   return((const char*)"Trace Frame-rate Scalability");
}


// ###### Check, if frame rate is scalable ##################################
bool TraceFrameRateScalability::isFrameRateScalable() const
{
   return(true);
}


// ###### Check, if frame rate is valid #####################################
bool TraceFrameRateScalability::isValidFrameRate(const double frameRate) const
{
   return(TraceReader->isValidFrameRate(frameRate));
}


// ###### Get nearest valid frame rate ######################################
double TraceFrameRateScalability::getNearestValidFrameRate(const double frameRate) const
{
   return(TraceReader->getNearestValidFrameRate(frameRate));
}


// ###### Get minimum frame rate ############################################
double TraceFrameRateScalability::getMinFrameRate() const
{
   return(TraceReader->getMinFrameRate());
}


// ###### Get maximum frame rate ############################################
double TraceFrameRateScalability::getMaxFrameRate() const
{
   return(TraceReader->getMaxFrameRate());
}


// ###### Get next frame rate ###############################################
double TraceFrameRateScalability::getNextFrameRateForRate(const double frameRate) const
{
   return(TraceReader->getNextFrameRateForRate(frameRate));
}


// ###### Get previous frame rate ###########################################
double TraceFrameRateScalability::getPrevFrameRateForRate(const double frameRate) const
{
   return(TraceReader->getPrevFrameRateForRate(frameRate));
}


// ###### Get frame rate scale factor #######################################
double TraceFrameRateScalability::getFrameRateScaleFactorForRate(const double frameRate) const
{
   const double min = TraceReader->getMinFrameRate();
   const double max = TraceReader->getMaxFrameRate();

   if(frameRate < min) {
      return(-HUGE_VAL);
   }
   else if(frameRate > max) {
      return(1.0);
   }

   if(min != max) {
      return((double)(frameRate - min) / (double)(max - min));
   }
   else {
      return(1.0);
   }
}


// ###### Get frame rate utilization ########################################
double TraceFrameRateScalability::getFrameRateUtilizationForRate(const double frameRate) const
{
      // ====== Get utilization header =========================================
   const UtilizationHeader* uh =
      TraceReader->getFrameRateUtilizationHeader(
         TraceReader->positionToFramePosition(frameRate,Position),
         frameRate);
   if(uh == NULL) {
      return(0.0);
   }

   const double scaleFactor = getFrameRateScaleFactorForRate(frameRate);
   return(evaluateUtilityFunctionTranslated(translate16(uh->Type),
                                            scaleFactor,
                                            (card64*)&uh->Constant,
                                            uh->Constants));
}


// ###### Get frame rate utilization ########################################
double TraceFrameRateScalability::getFrameRateUtilizationWeight(const double frameRate) const
{
   // ====== Get utilization header =========================================
   const UtilizationHeader* uh =
      TraceReader->getFrameRateUtilizationHeader(
         TraceReader->positionToFramePosition(frameRate,Position),
         frameRate);
   if(uh == NULL) {
      return(0.0);
   }

   return(translateToDouble(uh->Weight));
}


}
