// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Server                                                     ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "tdtfreader.h"
#include "socket.h"
#include "traceencoder.h"
#include "traceencoderrepository.h"
#include "rtpsender.h"
#include "rtcppacket.h"
#include "rtcpreceiver.h"
#include "rtcpabstractserver.h"
#include "mediainfo.h"
#include "sourcestateinfo.h"
#include "tools.h"
#include "randomizer.h"


#include "traceclientapppacket.h"
#include "traceserver.h"


// Verbose mode: Display some information on connections.
#define VERBOSE


// Debug mode: Print debug information
// #define BWMGR_DEBUG

// Test: Test of flow label changes
// #define FLOWLABEL_TEST

// Test: Test of traffic class changes
// #define TC_TEST



namespace Coral {


// ###### Constructor #######################################################
TraceServer::TraceServer(BandwidthManager* bwManager,
                         const cardinal    maxPacketSize)
{
   Randomizer random;
   OurSSRC      = random.random32();
   BandwidthMgr = bwManager;
   setMaxPacketSize(maxPacketSize);
   setLossScalability(true);
}


// ###### Destructor ########################################################
TraceServer::~TraceServer()
{
   stop();
}


// ###### Out of memory warning #############################################
void TraceServer::outOfMemoryWarning()
{
   printTimeStamp();
   cerr << "*** Out of memory! ***" << endl;
}


// ###### Send update to congestion manager #################################
void TraceServer::managementUpdate(const Client* client, User* user)
{
}


// ###### Add a new client ##################################################
void* TraceServer::newClient(Client* client, const char* cname)
{
   // ====== Create new user ================================================
   User* user = new User;
   if(user == NULL) {
      outOfMemoryWarning();
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }
   user->Client           = client;
   user->StreamIdentifier = 0;
   client->UserData       = user;


   // ====== Create and connect sender socket ===============================
   user->SenderSocket.create(Socket::IP,Socket::UDP);
   if(!user->SenderSocket.ready()) {
      cerr << "WARNING: TraceServer::newClient() - Unable to create socket!" << endl;
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }
   if(user->SenderSocket.bind() == false) {
      cerr << "WARNING: TraceServer::newClient() - Unable to bind socket!" << endl;
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }
   user->Flow = user->SenderSocket.allocFlow(client->ClientAddress);
   if(user->Flow.getFlowLabel() != 0) {
      if(user->SenderSocket.connect(user->Flow,TraceServerDefaultTrafficClass) == false) {
         if(user->SenderSocket.connect(client->ClientAddress,TraceServerDefaultTrafficClass) == false) {
            cerr << "WARNING: TraceServer::newClient() - Unable to flow-connect to client!" << endl;
            deleteClient(client,DeleteReason_Error);
            return(NULL);
         }
      }
   }
   else {
      if(user->SenderSocket.connect(client->ClientAddress,TraceServerDefaultTrafficClass) == false) {
         cerr << "WARNING: TraceServer::newClient() - Unable to connect to client!" << endl;
         deleteClient(client,DeleteReason_Error);
         return(NULL);
      }
   }


   // ====== Create repository and encoders =================================
   user->LastSequenceNumber  = 0xffff;
   user->PosChgSeqNumber     = 0xffff;
   user->ClientPause         = false;

   user->Repository.setAutoDelete(true);
   const bool e1 = user->Repository.addEncoder(new TraceEncoder(&user->Reader));
   if(e1 == false) {
      outOfMemoryWarning();
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }


   // ====== Initialize RTPSender ===========================================
   user->Sender.init(OurSSRC,&user->Repository,&user->SenderSocket,
                     MaxPacketSize,BandwidthMgr);


   // ====== Add stream to QoS management ===================================
   InternetAddress ourAddress;
   user->SenderSocket.getSocketAddress(ourAddress);
   UserSetSync.synchronized();
   user->StreamIdentifier = (integer)user;
   UserSet.insert(pair<const cardinal,User*>(user->StreamIdentifier,user));
   UserSetSync.unsynchronized();
/*
   if(BandwidthMgr != NULL) {
      BandwidthMgr->addStream(&user->Sender);
   }
*/

   // ====== Start RTPSender ================================================
   if(user->Sender.start() == false) {
      deleteClient(client,DeleteReason_Error);
      return(NULL);
   }


   // ====== Print information ==============================================
#ifdef VERBOSE
   printTimeStamp();
   char str[300];
   snprintf((char*)&str,sizeof(str),"New member $%08x added.",client->SSRC);
   cout << str << endl;
   InternetAddress sourceAddress;
   user->SenderSocket.getSocketAddress(sourceAddress);
   sourceAddress.setPrintFormat(InternetAddress::PF_Address);
   cout << "   CNAME:               " << cname << endl;
   cout << "   Source Address:      " << sourceAddress << endl;
   cout << "   Destination Address: " << (InternetAddress)client->ClientAddress << endl;
   if(user->SenderSocket.getSendFlowLabel() != 0) {
      snprintf((char*)&str,sizeof(str),"$%05x",user->SenderSocket.getSendFlowLabel());
      cout << "   Flow Label:          " << str << endl;
   }
   snprintf((char*)&str,sizeof(str),"$%02x",user->SenderSocket.getSendTrafficClass());
   cout << "   Traffic Class:       " << str << endl;
   cout << "   => We have " << getMembers() + 1 << " member(s) now!" << endl;
#endif
   return((void*)user);
}


// ###### Delete a client ###################################################
void TraceServer::deleteClient(Client* client, const DeleteReason reason)
{
   User* user = (User*)client->UserData;

   // ====== Remove user ====================================================
   if(user != NULL) {
      if(user->StreamIdentifier != 0) {
         UserSetSync.synchronized();
         multimap<const cardinal,User*>::iterator found =
            UserSet.find(user->StreamIdentifier);
         UserSet.erase(found);
         UserSetSync.unsynchronized();
      }
      if(BandwidthMgr != NULL) {
         BandwidthMgr->removeStream(&user->Sender);
      }
      user->Sender.stop();
      if(user->Flow.getFlowLabel() != 0) {
         user->SenderSocket.freeFlow(user->Flow);
      }
      delete user;
   }

   client->UserData = NULL;

   // ====== Print information ==============================================
#ifdef VERBOSE
   char str[128];
   printTimeStamp();
   snprintf((char*)&str,sizeof(str),"$%08x",client->SSRC);

   cout << str << " removed";
   switch(reason) {
      case DeleteReason_Timeout:
         cout << " due to timeout!" << endl;
       break;
      case DeleteReason_Shutdown:
         cout << " due to server shutdown!" << endl;
        break;
      case DeleteReason_Error:
         cout << " due to transmission error!" << endl;
        break;
      default:
         cout << "." << endl;
        break;
   }
   if(reason != DeleteReason_Shutdown) {
      cout << "   => We have " << getMembers() - 1 << " member(s) now!" << endl;
   }
#endif
}


// ###### Handle user command ###############################################
void TraceServer::userCommand(const Client*               client,
                              User*                       user,
                              const TraceClientAppPacket* app)
{
   // ====== Check, if command has already been executed ====================
   const integer diff = (integer)app->SequenceNumber -
                           (integer)user->LastSequenceNumber;
   if(!((diff > 0) || (diff < -30000))) {
      return;
   }

   user->LastSequenceNumber = app->SequenceNumber;
   user->Sender.synchronized();

   // ====== Set quality, position, media name and bandwidth limit ==========
   if(user->Repository.selectEncoderForTypeID(app->Encoding) == true) {
      // ====== Update user's settings ======================================
      user->Repository.setWantedUtilization(translateToDouble(app->WantedUtilization));
      user->Repository.setStreamPriority(app->StreamPriority);
      user->Repository.setMaxWantedDelay((double)app->MaxWantedDelay);
      user->Repository.setMinWantedBandwidth(app->MinWantedBandwidth);
      user->Repository.setMaxWantedBandwidth(app->MaxWantedBandwidth);


      // ====== Select new media and set frame number =======================
      if(user->Reader.getErrorCode() == ME_NoMedia) {
#ifdef VERBOSE
         char str[128];
         printTimeStamp();
         snprintf((char*)&str,sizeof(str),"$%08x",client->SSRC);
         cout << str << " loading media <" << (const char*)&app->MediaName
              << ">." << endl;
#endif
         if(user->Reader.open((char*)&app->MediaName,false,true)) {
            user->Reader.setPosition(app->RestartPosition);
            user->PosChgSeqNumber = app->PosChgSeqNumber;
            user->MediaName       = String((const char*)&app->MediaName);
            user->Sender.leaveCorrectionLoop();
            if(BandwidthMgr != NULL) {
               BandwidthMgr->addStream(&user->Sender);
            }
         }
      }
      else if((user->MediaName != "") && (user->MediaName != (char*)&app->MediaName)) {
#ifdef VERBOSE
         char str[128];
         printTimeStamp();
         snprintf((char*)&str,sizeof(str),"$%08x",client->SSRC);
         cout << str << " changing media to <" << (const char*)&app->MediaName
              << ">." << endl;
#endif

         if(BandwidthMgr != NULL) {
            BandwidthMgr->removeStream(&user->Sender);
         }
         user->Reader.close();
         if(user->Reader.open((char*)&app->MediaName,false,true)) {
            user->Reader.setPosition(app->RestartPosition);
            user->MediaName = String((const char*)&app->MediaName);
            user->Sender.leaveCorrectionLoop();

            if(BandwidthMgr != NULL) {
               BandwidthMgr->addStream(&user->Sender);
            }
         }
      }


      // ====== Change position, if it has been updated =====================
      const integer diff = (integer)app->PosChgSeqNumber -
                              (integer)user->PosChgSeqNumber;
      if((diff > 0) || (diff < -30000)) {
         user->PosChgSeqNumber = app->PosChgSeqNumber;
         if(app->StartPosition != (card64)-1) {
            user->Reader.setPosition(app->StartPosition);
            user->Sender.leaveCorrectionLoop();
         }
      }
   }
   else {
      char str[32];
      snprintf((char*)&str,sizeof(str),"$%04x",app->Encoding);
      cerr << "WARNING: TraceServer::userCommand() - Unsupported Encoding #"
           << str << " requested!" << endl;
   }


   // ====== Execute command ================================================
   switch(app->Status) {
      case TraceClientAppPacket::TCAS_Pause:
        user->ClientPause = true;
        user->Sender.setPause(true);
       break;
      case TraceClientAppPacket::TCAS_Play:
         user->ClientPause = false;
         user->Sender.setPause(false);
        break;
   }

   user->Sender.unsynchronized();
}


// ###### Handle app message ################################################
void TraceServer::appMessage(const Client*  client,
                             const char*    name,
                             const void*    data,
                             const cardinal dataLength)
{
   User* user = (User*)client->UserData;
   TraceClientAppPacket* app = (TraceClientAppPacket*)data;
   app->translate();
   if(app->FormatID == TraceClientAppPacket::TraceClientFormatID) {
      userCommand(client,user,app);
      managementUpdate(client,user);
   }
}


// ###### Handle SDES message ###############################################
void TraceServer::sdesMessage(const Client*  client,
                              const card8    type,
                              const char*    data,
                              const cardinal length)
{
   User* user = (User*)client->UserData;
   if(type == RTCP_SDES_PRIV) {
      TraceClientAppPacket* app = (TraceClientAppPacket*)data;
      app->translate();
      if(app->FormatID == TraceClientAppPacket::TraceClientFormatID) {
         userCommand(client,user,app);
         managementUpdate(client,user);
      }
   }
/*
   else {
      cerr << "NOTE: TraceServer::sdesMessage() - Unsupported SDES type: "
           << (cardinal)type << endl;
   }
*/
}


// ###### Check, if client is okay ##########################################
bool TraceServer::checkClient(const Client* client)
{
   User* user = (User*)client->UserData;

   return(user->Sender.transmissionErrorDetected() == false);
}


// ###### Handle receiver report ############################################
void TraceServer::receiverReport(const Client*                   client,
                                 const RTCPReceptionReportBlock* report,
                                 const cardinal                  layer)
{
   User* user = (User*)client->UserData;

   if(LossScalability == true) {
      if(BandwidthMgr != NULL) {
#ifdef BWMGR_DEBUG
         cout << "Loss rate in layer #" << layer << ": " << report->getFractionLost() << endl;
#endif
         BandwidthMgr->lossEvent(&user->Sender,layer,report->getFractionLost());
      }
   }
}


}
