// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Service Level Agreement Implementation                           ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "servicelevelagreement.h"


namespace Coral {


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const ServiceLevelAgreement sla)
{
   char str[256];
   os << "Idx  TC    Bandwidth  Cost Factor    MaxDelay [ms]  MaxLossRate    MaxJitter [ms]"
      << endl;
   os << "---------------------------------------------------------------------------------"
      << endl;
   for(cardinal i = 0;i < sla.Classes;i++) {
      snprintf((char*)&str,sizeof(str),
              "#%02d  %-4s  %7d    %1.2f           %1.0f            %-1.2f           %-1.3f",
              i,
              TrafficClassValues::getNameForTrafficClass(sla.Class[i].TrafficClass),
              sla.Class[i].BytesPerSecond,
              sla.Class[i].CostFactor,
              sla.Class[i].MaxTransferDelay / 1000.0,
              sla.Class[i].MaxLossRate,
              sla.Class[i].MaxJitter);
      os << str << endl;
   }
   os << "---------------------------------------------------------------------------------"
      << endl
      << "Total " << sla.TotalBandwidth << endl;
   return(os);
};


// ###### Constructor #######################################################
ServiceLevelAgreement::ServiceLevelAgreement()
{
   Classes        = 0;
   TotalBandwidth = 0;
}


// ###### Destructor ########################################################
ServiceLevelAgreement::~ServiceLevelAgreement()
{
}


// ###### Load configuration from file ######################################
bool ServiceLevelAgreement::load(const char* fileName)
{
   // ====== Open file ======================================================
   FILE* inputFD = fopen(fileName,"r");
   if(inputFD == NULL) {
      cerr << "ERROR: Unable to open input file <" << fileName << ">!" << endl;
      return(false);
   }

   // ====== Read configuration from file ===================================
   char str[256];
   char* result = fgets((char*)&str,256,inputFD);
   cardinal line  = 0;
   Classes        = 0;
   TotalBandwidth = 0;
   while(!feof(inputFD)) {
      line++;
      const cardinal inputLength = strlen((char*)&str);
      if(inputLength > 1) {
         str[inputLength - 1] = 0x00;
         switch(str[0]) {
            // ====== Line is a comment =====================================
            case '#':
             break;

            // ====== Line is a setting =====================================
            default:
               const String input(str);
               String name;
               String value;
               if(input.scanSetting(name,value)) {
                  if(name == "CLASS") {
                     if(Classes >= TrafficClassValues::MaxValues) {
                        cerr << "ERROR: Too many classes!" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     unsigned int bandwidth;
                     double       maxJitter;
                     double       maxLossRate;
                     double       maxDelay;
                     double       costFactor;
                     char         name[20];
                     cardinal count =
                        sscanf(value.getData(),
                               "%5s %d %lf %lf %lf %lf",
                               (char*)&name,
                               &bandwidth,
                               &costFactor,
                               &maxDelay,
                               &maxLossRate,
                               &maxJitter);
                     if(count < 6) {
                        cerr << "ERROR: Bad class setting, "
                                "line " << line << "!" << endl;
                        cerr << "       Syntax: Class = <name> <bandwidth> <cost factor> <max delay> <max loss> <max jitter>" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     const card16 tc = TrafficClassValues::getTrafficClassForName(name);
                     if(tc == 0xffff) {
                        cerr << "ERROR: Unknown class name, "
                                "line " << line << "!" << endl;
                        fclose(inputFD);
                        return(false);
                     }
                     if(maxLossRate < 0.0) {
                        maxLossRate = 1.0;
                     }
                     else if(maxLossRate > 1.0) {
                        maxLossRate = 1.0;
                     }
                     if(maxJitter < 0.0) {
                        maxJitter = HUGE_VAL;
                     }
                     if(maxDelay < 0.0) {
                        maxDelay = HUGE_VAL;
                     }
                     if(costFactor < 0.0) {
                        costFactor = 0.0;
                     }
                     Class[Classes].BytesPerSecond   = bandwidth;
                     Class[Classes].MaxTransferDelay = maxDelay;
                     Class[Classes].MaxLossRate      = maxLossRate;
                     Class[Classes].MaxJitter        = maxJitter;
                     Class[Classes].CostFactor       = costFactor;
                     Class[Classes].TrafficClass     = (card8)tc;
                     Classes++;
                     TotalBandwidth += bandwidth;
                  }
                  else {
                     cerr << "ERROR: Unknown option <"
                          << name << " = " << value << ">, "
                             "line " << line << "!" << endl;
                     fclose(inputFD);
                     return(false);
                  }
               }
             break;
         }
      }
      result = fgets((char*)&str,256,inputFD);
   }
   fclose(inputFD);


   // ====== Sort classes by cost ascending =================================
   for(cardinal i = 0;i < Classes;i++) {
      for(cardinal j = i + 1;j < Classes;j++) {
         if(Class[i].CostFactor > Class[j].CostFactor) {
            const DiffServClass tmp = Class[i];
            Class[i] = Class[j];
            Class[j] = tmp;
         }
      }
   }
   return(true);
}


// ###### Get possible classes for BandwidthInfo ############################
cardinal ServiceLevelAgreement::getPossibleClassesForBandwidthInfo(
                                   const BandwidthInfo& bandwidthInfo,
                                   cardinal*            classList) const
{
   cardinal count = 0;
   for(cardinal i = 0;i < Classes;i++) {
      if(Class[i] >= bandwidthInfo) {
         classList[count] = i;
         count++;
      }
   }
   return(count);
}


}
