// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Encoder Repository Implementation                          ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "traceencoderrepository.h"


namespace Coral {


// ###### Constructor #######################################################
TraceEncoderRepository::TraceEncoderRepository()
{
   Encoder = NULL;
}


// ###### Destructor ########################################################
TraceEncoderRepository::~TraceEncoderRepository()
{
   if(Encoder) {
      Encoder->deactivate();
   }
   if(AutoDelete) {
      while(TraceEncoderRepository.begin() != TraceEncoderRepository.end()) {
         multimap<const card16,TraceEncoderInterface*>::iterator encoderIterator =
            TraceEncoderRepository.begin();
         Encoder = encoderIterator->second;
         TraceEncoderRepository.erase(encoderIterator);
         delete Encoder;
      }
   }
   Encoder = NULL;
}


// ###### Add encoder to repository #########################################
bool TraceEncoderRepository::addEncoder(TraceEncoderInterface* encoder)
{
   multimap<const card16,TraceEncoderInterface*>::iterator encoderIterator =
      TraceEncoderRepository.find(encoder->getTypeID());
   if(encoderIterator == TraceEncoderRepository.end()) {
      TraceEncoderRepository.insert(pair<const card16,TraceEncoderInterface*>
                                    (encoder->getTypeID(),encoder));
      if(Encoder == NULL) {
         Encoder = encoder;
      }
      return(true);
   }
   return(false);
}


// ###### Remove encoder from repository ####################################
void TraceEncoderRepository::removeEncoder(TraceEncoderInterface* encoder)
{
   multimap<const card16,TraceEncoderInterface*>::iterator encoderIterator =
      TraceEncoderRepository.find(encoder->getTypeID());
   if(encoderIterator != TraceEncoderRepository.end()) {
      TraceEncoderRepository.erase(encoderIterator);
      encoder->deactivate();
      if(Encoder == encoder) {
         multimap<const card16,TraceEncoderInterface*>::iterator firstEncoder =
            TraceEncoderRepository.begin();
         if(firstEncoder != TraceEncoderRepository.end()) {
            Encoder = firstEncoder->second;
         }
         else {
            Encoder = NULL;
         }
      }
      else {
         Encoder = NULL;
      }
   }
}


// ###### Select encoder for given type ID ##################################
bool TraceEncoderRepository::selectEncoderForTypeID(const card16 typeID)
{
   if((Encoder == NULL) || (typeID != Encoder->getTypeID())) {
      multimap<const card16,TraceEncoderInterface*>::iterator encoderIterator =
         TraceEncoderRepository.find(typeID);
      if(encoderIterator != TraceEncoderRepository.end()) {
          TraceEncoderInterface* encoder = encoderIterator->second;
          Encoder->deactivate();
          const double utilization = Encoder->getWantedUtilization();

          Encoder = encoder;

          Encoder->setWantedUtilization(utilization);
          Encoder->activate();
          return(true);
      }
      return(false);
   }
   return(true);
}


// ###### Get pointer to current encoder ####################################
EncoderInterface* TraceEncoderRepository::getCurrentEncoder() const
{
   return((EncoderInterface*)Encoder);
}
   

// ###### Get pointer to current trace encoder ##############################
TraceEncoderInterface* TraceEncoderRepository::getCurrentTraceEncoder() const
{
   return(Encoder);
}



/* 
#############################################################################
###### State Pattern: Method calls                                     ######
#############################################################################
*/

const card16 TraceEncoderRepository::getTypeID() const
   { return(Encoder->getTypeID()); }

const char* TraceEncoderRepository::getTypeName() const
   { return(Encoder->getTypeName()); }

void TraceEncoderRepository::activate()
   { Encoder->activate(); }

void TraceEncoderRepository::deactivate()
   { Encoder->deactivate(); }

void TraceEncoderRepository::reset()
   { Encoder->reset(); }
   
bool TraceEncoderRepository::prepareNextFrame(const cardinal headerSize,
                                              const cardinal maxPacketSize,
                                              const cardinal flags)
   { return(Encoder->prepareNextFrame(headerSize,maxPacketSize,flags)); }

cardinal TraceEncoderRepository::getNextPacket(EncoderPacket* encoderPacket)
   { return(Encoder->getNextPacket(encoderPacket)); }

double TraceEncoderRepository::getWantedUtilization() const
   { return(Encoder->getWantedUtilization()); }

void TraceEncoderRepository::setWantedUtilization(const double utilization)
   { Encoder->setWantedUtilization(utilization); }

double TraceEncoderRepository::getMaxWantedDelay() const
   { return(Encoder->getMaxWantedDelay()); }

void TraceEncoderRepository::setMaxWantedDelay(const double delay)
   { Encoder->setMaxWantedDelay(delay); }

cardinal TraceEncoderRepository::getMinWantedBandwidth() const
   { return(Encoder->getMinWantedBandwidth()); }

cardinal TraceEncoderRepository::getMaxWantedBandwidth() const
   { return(Encoder->getMaxWantedBandwidth()); }

void TraceEncoderRepository::setMinWantedBandwidth(const cardinal bandwidth)
   { Encoder->setMinWantedBandwidth(bandwidth); }

void TraceEncoderRepository::setMaxWantedBandwidth(const cardinal bandwidth)
   { Encoder->setMaxWantedBandwidth(bandwidth); }

int8 TraceEncoderRepository::getStreamPriority() const
   { return(Encoder->getStreamPriority()); }

void TraceEncoderRepository::setStreamPriority(const int8 priority)
   { Encoder->setStreamPriority(priority); }

AbstractQoSDescription* TraceEncoderRepository::getQoSDescription(
                           const cardinal pktHeaderSize,
                           const cardinal pktMaxSize,
                           const card64   position)
   { return(Encoder->getQoSDescription(pktHeaderSize,pktMaxSize,position)); }

void TraceEncoderRepository::updateQuality(const AbstractQoSDescription* aqd)
   { Encoder->updateQuality(aqd); }

card64 TraceEncoderRepository::checkInterval()
   { return(Encoder->checkInterval()); }


}
