// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace QoS Description Implementation                             ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "traceqosdescription.h"


namespace Coral {



// ##########################################################################
// #### Trace Layer QoS Description                                      ####
// ##########################################################################



// ###### Get frame size scalability class name #############################
const char* TraceLayerDescription::getFrameSizeScalabilityClass() const
{
   return((const char*)"Trace Frame-size Scalability");
}


// ##########################################################################
// #### Trace QoS Description                                            ####
// ##########################################################################


// ###### Initialize Trace QoS description ##################################
void TraceQoSDescription::initTraceDescription(TDTFReader*  traceReader,
                                               const card64 position,
                                               const card64 maxPosition,
                                               const double frameRate)
{
   setPosition(position);
   TraceReader = traceReader;
   Position    = position;
   MaxPosition = maxPosition;
   FrameRate   = frameRate;
}


// ###### Update Trace QoS description #######################################
void TraceQoSDescription::updateDescription(const cardinal pktHeaderSize,
                                            const cardinal pktMaxSize)
{
   // ====== Initialize FrameRateScalability object =========================
   initFrameRateScalability(TraceReader,Position);

   // ====== Initialize AbstractQoSDescription ==============================
   initDescription(FrameRate);

   // ====== Initialize layers ==============================================
   const cardinal layers =
      min(MaxLayers,TraceReader->getLayers(FrameRate));

   for(cardinal i = 0;i < layers;i++) {
      Layer[i].initLayer(
         pktHeaderSize, pktMaxSize,
         (cardinal)-1,
         1000.0, 20,
         0.15,
         HUGE_VAL,
         TraceReader->getLayerFlags(Position,i,FrameRate));
      Layer[i].initFrameSizeScalability(
                  TraceReader, Position,
                  i, TraceReader->getLayerScalability(Position,i,FrameRate));
   }
}


// ###### Get number of layers ##############################################
cardinal TraceQoSDescription::getLayers() const
{
   return(min(MaxLayers,TraceReader->getLayers(FrameRate)));
}


// ###### Get layer #########################################################
AbstractLayerDescription* TraceQoSDescription::getLayer(const cardinal layer) const
{
   if(layer >= getLayers()) {
      cerr << "ERROR: AbstractQoSDescription::getLayer() - " << endl
           << "Invalid parameter " << layer << "!" << endl;
   }
   return((AbstractLayerDescription*)&Layer[layer]);
}


// ###### Get frame rate scalability class name #############################
const char* TraceQoSDescription::getFrameRateScalabilityClass() const
{
   return((const char*)"Trace Frame-rate Scalability");
}


// ###### Recursive ResourceUtilizationPoint calculation method #############
void TraceQoSDescription::doSelectIteration(
                             bool*          select,
                             const cardinal bwThreshold,
                             const double   utThreshold,
                             double*        utilizationCache,
                             cardinal*      bandwidthCache,
                             const cardinal maxPoints,
                             const cardinal maxCachePoints,
                             const cardinal start,
                             const cardinal end,
                             const cardinal startBandwidth,
                             const cardinal endBandwidth,
                             const cardinal level,
                             const cardinal maxLevel,
                             cardinal&      count) const
{
   // ====== Calculate middle cache position and bandwidth ==================
   const cardinal m            = (start + end) / 2;
   const cardinal newBandwidth = (startBandwidth + endBandwidth) / 2;


   // ====== If level is lowest: Calculate new ResourceUtilizationPoint =====
   if(level == maxLevel) {
      // ====== Add point, if utilization and bandwidth is useful ===========
      if(utilizationCache[m] >= 0.0) {
         // ====== Get nearest lower point's utilization ====================
         double uLeft = utilizationCache[m];
         for(integer i = (integer)m - 1;i >= 0;i--) {
            if(utilizationCache[i] != -HUGE_VAL) {
               uLeft = utilizationCache[i];
               break;
            }
         }

         // ====== Get nearest higher point's utilization ===================
         double uRight = utilizationCache[m];
         for(cardinal i = m + 1;i < maxCachePoints;i++) {
            if(utilizationCache[i] != -HUGE_VAL) {
               uRight = utilizationCache[i];
               break;
            }
         }


         // Add point, if utilization or bandwidth difference is greater than
         // threshold.
         if( ((fabs(utilizationCache[m] - uLeft) >= utThreshold) &&
              (fabs(uRight - utilizationCache[m]) >= utThreshold)) ||
             ((bandwidthCache[m] >= bandwidthCache[start] + bwThreshold) &&
              (bandwidthCache[m] + bwThreshold < bandwidthCache[end])) ) {
            select[m] = true;
            count++;
         }
         else {
            utilizationCache[m] = -HUGE_VAL;
         }
      }
      else {
         utilizationCache[m] = -HUGE_VAL;
      }
   }


   // ====== Go into next recursion level ===================================
   else {
      if(count < maxPoints) {
         if(start + 1 < m) {
            doSelectIteration(select,
                              bwThreshold, utThreshold,
                              utilizationCache, bandwidthCache,
                              maxPoints,
                              maxCachePoints,
                              start, m,
                              startBandwidth, newBandwidth,
                              level + 1, maxLevel, count);
         }
      }
      if(count < maxPoints) {
         if(m + 1 < end) {
            doSelectIteration(select,
                              bwThreshold, utThreshold,
                              utilizationCache, bandwidthCache,
                              maxPoints,
                              maxCachePoints,
                              m, end,
                              newBandwidth, endBandwidth,
                              level + 1, maxLevel, count);
         }
      }
   }
}


// ###### Calculate utilization for layer bandwidths ########################
double TraceQoSDescription::calculateUtilizationForLayerBandwidths(
          const double    frameRate,
          const cardinal  layers,
          const cardinal* bandwidth) const
{
   // ====== Check for end of file ==========================================
   if(Position >= MaxPosition) {
      return(1.0);
   }

   // ====== Do standard utilization calculation ============================
   return(AbstractQoSDescription::calculateUtilizationForLayerBandwidths(
             frameRate,layers,bandwidth));
}


// ###### Get precomputed resource/utilization list #########################
cardinal TraceQoSDescription::getPrecomputedResourceUtilizationList(
                                 ResourceUtilizationPoint* rup,
                                 const cardinal            bwThreshold,
                                 const double              utThreshold,
                                 const cardinal            maxPoints) const
{
   // ====== Check for end of file ==========================================
   if(Position >= MaxPosition) {
      return(0);
   }


   // ====== Get resource/utilization header ================================
   const ResourceUtilizationHeader* ruh =
      TraceReader->getResourceUtilizationHeader(
         TraceReader->positionToFramePosition(TraceReader->getMaxFrameRate(),
                                              Position),
         TraceReader->getMaxFrameRate());
   if(ruh == NULL) {
      cerr << "ERROR: TraceQoSDescription::getPrecomputedResourceUtilizationList() - "
              "Missing resource/utilization header!" << endl;
      exit(1);
   }
   const cardinal entries = translate16(ruh->Entries);
   if(entries == 0) {
      cerr << "WARNING: TraceQoSDescription::getPrecomputedResourceUtilizationList() - "
           << "Empty resource/utilization list!" << endl;
      return(0);
   }
   const cardinal layers = getLayers();


   // ====== Prepare for recursive algorithm ================================
   double   utilizationCache[entries];
   cardinal bandwidthCache[entries];
   bool     select[entries];
   const cardinal iterations = (cardinal)ceil(log((double)entries) / log(2.0));
   const ResourceUtilizationEntry* entry = &ruh->Entry[0];
   for(cardinal i = 0;i < entries;i++) {
      utilizationCache[i] = (double)translate32(entry->Utilization) / (double)pow(2.0,30.0);
      bandwidthCache[i]   = translate32(entry->Bandwidth[0]);
      select[i]           = false;
      entry = (ResourceUtilizationEntry*)
                 ((long)entry +
                  (long)(sizeof(ResourceUtilizationEntry) +
                     ((layers + 1) * sizeof(card32))));
   }
   const cardinal minBandwidth = bandwidthCache[0];
   const cardinal maxBandwidth = bandwidthCache[entries - 1];
   select[0]           = true;
   select[entries - 1] = true;


   // ====== Execute recursive algorithm ====================================
   cardinal count = 2;
   for(cardinal i = 0;count < maxPoints;i++) {
      doSelectIteration(select,
         bwThreshold, utThreshold,
         (double*)&utilizationCache, (cardinal*)&bandwidthCache,
         maxPoints,
         entries,
         0, entries - 1,
         minBandwidth, maxBandwidth,
         0, i, count);

      if(i >= iterations) {
         break;
      }
   }


   // ====== Copy selected points ===========================================
   count = 0;
   entry = &ruh->Entry[0];
   for(cardinal i = 0;i < entries;i++) {
      if(select[i] == true) {
         rup[count].Utilization = utilizationCache[i];
         rup[count].FrameRate   = translateToDouble(entry->FrameRate);
         rup[count].Bandwidth   = 0;
         rup[count].Layers      = layers;
         rup[count].TaskID      = 0;
         rup[count].Cost        = 0.0;
         for(cardinal j = 0;j < layers;j++) {

            const cardinal payloadFS = Layer[j].bandwidthToFrameSize(rup[count].FrameRate,translate32(entry->Bandwidth[j + 1]));
            const cardinal rawFS     = Layer[j].payloadToRaw(rup[count].FrameRate,payloadFS);
            const cardinal bw        = Layer[j].frameSizeToBandwidth(rup[count].FrameRate,rawFS);

            rup[count].LayerBandwidthInfo[j].BytesPerSecond = bw;
            rup[count].Bandwidth += bw;
            rup[count].LayerBandwidthInfo[j].PacketsPerSecond =
               Layer[j].frameSizeToPacketRate(
                  rup[count].FrameRate,
                  Layer[j].bandwidthToFrameSize(rup[count].FrameRate,rup[count].LayerBandwidthInfo[j].BytesPerSecond)
               );

            rup[count].LayerBandwidthInfo[j].BufferDelay      = Layer[j].getBufferDelay();
            rup[count].LayerBandwidthInfo[j].MaxLossRate      = Layer[j].getMaxLossRate();
            rup[count].LayerBandwidthInfo[j].MaxTransferDelay = Layer[j].getMaxTransferDelay();
            rup[count].LayerBandwidthInfo[j].MaxJitter        = Layer[j].getMaxJitter();
         }
         count++;
      }
      entry = (ResourceUtilizationEntry*)
              ((long)entry +
               (long)(sizeof(ResourceUtilizationEntry) +
                  ((layers + 1) * sizeof(card32))));
   }

   return(count);
}


}
