// ##########################################################################
// ####                                                                  ####
// ####                    Master Thesis Implementation                  ####
// ####  Management of Layered Variable Bitrate Multimedia Streams over  ####
// ####                  DiffServ with A Priori Knowledge                ####
// ####                                                                  ####
// #### ================================================================ ####
// ####                                                                  ####
// ####                                                                  ####
// #### Trace Utilizer                                                   ####
// ####                                                                  ####
// #### Version 1.00  --  October 10, 2000                               ####
// ####                                                                  ####
// #### Copyright (C) 2000 Thomas Dreibholz                              ####
// #### University of Bonn, Department of Computer Science IV            ####
// #### EMail: Dreibholz@bigfoot.com                                     ####
// #### WWW:   http://www.bigfoot.com/~dreibholz/diplom/index.html       ####
// ####                                                                  ####
// ##########################################################################


#include "system.h"
#include "tdtf.h"
#include "tdtfmediareader.h"
#include "mpegwriterqosdescription.h"
#include "mp3writerqosdescription.h"
#include "globaltraceconfiguration.h"


using namespace Coral;


/**
  * This class extends the TDTF media reader by a method to do
  * utilization calculations.
  *
  * @short   TDTF Media Reader.
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class TDTFUtilizationUpdater : public TDTFMediaReader
{
   // ====== Calculate utilization ==========================================
   public:
   /**
     * Write new utilization information to TDTF file represented by
     * TDTFMediaReader. After the calculation, the file will be closed!
     *
     * @param qosDescription QoS description for the trace's media type.
     * @param config TraceConfiguration.
     * @return true, if operation was successful; false otherwise.
     */
   bool doUpdate(TraceQoSDescription*      qosDescription,
                 const TraceConfiguration& config,
                 const bool                truncateOnly);
};


// ###### Update all intervals of given frame rate ##########################
bool TDTFUtilizationUpdater::doUpdate(TraceQoSDescription*      qosDescription,
                                      const TraceConfiguration& config,
                                      const bool                truncateOnly)
{
   // ====== Count frame rates ==============================================
   if(InputMemory == NULL) {
      close();
      return(false);
   }
   cardinal frameRates = 0;
   double fr = getMinFrameRate();
   while(fr != getMaxFrameRate()) {
      fr = getNextFrameRateForRate(fr);
      frameRates++;
   }
   frameRates++;


   // ====== Count intervals ================================================
   cardinal intervals[frameRates];
   cardinal frames[frameRates];
   double   frameRate[frameRates];
   fr = getMinFrameRate();
   for(cardinal i = 0;i < frameRates;i++) {
      cardinal position  = 0;
      cardinal length    = 0;
      intervals[i] = 0;
      frames[i]    = getFrames(fr);
      frameRate[i] = fr;
      do {
         // ====== Go to next interval ======================================
         const IntervalHeader* ih = getIntervalHeader(position,fr);
         if(ih == NULL) {
            cerr << "ERROR: update() - Interval header not found!" << endl;
            close();
            return(false);
         }
         length = translate32(ih->Length);
         position += length;
         intervals[i]++;
      } while(position < frames[i]);
      fr = getNextFrameRateForRate(fr);
   }
   const cardinal maxRateIndex = frameRates - 1;


   // ====== Reset utilization calculation time stamps ======================
   TDTFPrefix* tp               = (TDTFPrefix*)getTDTFPrefix();
   const TDTFSuffix* originalTS = (TDTFSuffix*)getTDTFSuffix();
   TDTFSuffix  ts = *originalTS;
   tp->UtilizationTimeStamp = 0;
   ts.UtilizationTimeStamp  = 0;
   const card64 startTS = getMicroTime();


   // ====== Truncate old utilization block =================================
   cardinal filePosition = translate32(ts.ResourceUtilizationStart);
   if(ftruncate(InputFile,(off_t)filePosition) != 0) {
      cerr << "ERROR: update() - File truncation failed!" << endl;
      close();
      return(false);
   }
   if(lseek(InputFile,(off_t)filePosition,SEEK_SET) != (off_t)filePosition) {
      cerr << "ERROR: update() - Seek failed!" << endl;
   }


   // ====== Write resource/utilization lists ===============================
   cardinal ruIndex = 0;
   if(!truncateOnly) {
      IntervalHeader* ih[frameRates];
      IntervalHeader* oldIH[frameRates];
      cardinal        length[frameRates];
      for(cardinal i = 0;i < frameRates;i++) {
         oldIH[i] = NULL;
      }
      const cardinal maxIndexEntries = frames[maxRateIndex];
      ResourceUtilizationListIndexEntry index[maxIndexEntries];
      cardinal indexPosition = 0;

      for(cardinal fposition = 0;fposition < frames[maxRateIndex];fposition++) {
         const card64 position = framePositionToPosition(frameRate[maxRateIndex],fposition);

         // ====== Get interval headers for new position, check for changes =
         bool changed = false;
         for(cardinal k = 0;k < frameRates;k++) {
            // ====== Ger interval header for each frame rate ===============
            const cardinal framePosition =
               positionToFramePosition(frameRate[k],position);
            if(framePosition < frames[k]) {
               ih[k] = (IntervalHeader*)getIntervalHeader(
                          framePosition,
                          frameRate[k]);
               if(ih[k] == NULL) {
                  cerr << "ERROR: update() - Interval header not found!" << endl;
                  close();
                  return(false);
               }
               length[k] = translate32(ih[k]->Length);

               // ====== Check, if interval is new ==========================
               if(ih[k] != oldIH[k]) {
                  changed = true;
                  oldIH[k] = ih[k];


                  // ====== Update frame rate utilization header ============
                  UtilizationHeader* uh =
                     (UtilizationHeader*)getFrameRateUtilizationHeader(framePosition,frameRate[k]);
                  if(uh == NULL) {
                     cerr << "ERROR: update() - Frame rate utilization header not found!" << endl;
                     close();
                     return(false);
                  }
                  if((uh->MaxConstants < config.FrameRateUtilizationMaxConstants) ||
                     (uh->MaxConstants < config.FrameRateUtilizationConstants)) {
                     cerr << "ERROR: update() - Frame rate utilization constant space too low!" << endl;
                     close();
                     return(false);
                  }
                  uh->Constants = config.FrameRateUtilizationConstants;
                  uh->Type      = translate16(config.FrameRateUtilizationType);
                  uh->Weight    = translateToBinary(config.FrameRateUtilizationWeight);
                  for(cardinal i = config.FrameRateUtilizationConstants;i < uh->MaxConstants;i++) {
                     uh->Constant[i] = translateToBinary(0.0);
                  }
                  for(cardinal i = 0;i < config.FrameRateUtilizationConstants;i++) {
                     uh->Constant[i] = translateToBinary(config.FrameRateUtilizationConstant[i]);
                  }


                  // ====== Update frame size utilization header ============
                  for(cardinal i = 0;i < ih[k]->Layers;i++) {
                     uh = (UtilizationHeader*)getFrameSizeUtilizationHeader(framePosition,i,frameRate[k]);
                     if(uh == NULL) {
                        cerr << "ERROR: update() - Frame size utilization header not found!" << endl;
                        close();
                        return(false);
                     }
                     if((uh->MaxConstants < config.FrameRateUtilizationMaxConstants) ||
                        (uh->MaxConstants < config.FrameRateUtilizationConstants)) {
                        cerr << "ERROR: update() - Frame size utilization constant space too low!" << endl;
                        close();
                        return(false);
                     }
                     uh->Constants = config.Layer[i].FrameSizeUtilizationConstants;
                     uh->Type      = translate16(config.Layer[i].FrameSizeUtilizationType);
                     uh->Weight    = translateToBinary(config.Layer[i].FrameSizeUtilizationWeight);
                     for(cardinal j = config.Layer[i].FrameSizeUtilizationConstants;j < uh->MaxConstants;j++) {
                        uh->Constant[j] = translateToBinary(0.0);
                     }
                     for(cardinal j = 0;j < config.Layer[i].FrameSizeUtilizationConstants;j++) {
                        uh->Constant[j] = translateToBinary(config.Layer[i].FrameSizeUtilizationConstant[j]);
                     }
                  }
               }
            }
         }


         // ====== Write new list, if change has been detected ==============
         if(changed == true) {
            // ====== Add index entry =======================================
            if(indexPosition > 0) {
               index[indexPosition - 1].Length =
                  translate32(fposition - translate32(index[indexPosition - 1].Position));
            }
            index[indexPosition].Position = translate32(fposition);
            index[indexPosition].Length   = 0;
            index[indexPosition].List     = translate32(filePosition);
            indexPosition++;


            // ====== Set frame rate and position ===========================
            qosDescription->initTraceDescription(
               this, position,
               getMaxPosition(),
               frameRate[maxRateIndex]);


            // NOTE: We want to use *payload* here -> set header size to 0!
            qosDescription->updateDescription(0,1024);


            // ====== Calculate resource/utilization list ===================
            ResourceUtilizationPoint rupList[config.RUPoints];
            const cardinal newEntries =
               qosDescription->calculateResourceUtilizationList(
                  (ResourceUtilizationPoint*)&rupList,
                  config.BandwidthThreshold,
                  config.UtilizationThreshold,
                  config.RUPoints);
            if(newEntries < 2) {
               cerr << "ERROR: TDTFUtilizationUpdater::doUpdate() - Too few points calculated?!" << endl;
               exit(1);
            }
            ResourceUtilizationPoint::sortResourceUtilizationList(
                                         (ResourceUtilizationPoint*)&rupList,
                                         0, newEntries - 1);


            // ====== Write resource/utilization list =======================
            char ruListBuffer[ResourceUtilizationHeader::getResourceUtilizationSize(ih[maxRateIndex]->Layers,newEntries)];
            ResourceUtilizationHeader* ruh = (ResourceUtilizationHeader*)&ruListBuffer;
            ruh->Entries  = translate16(newEntries);
            ruh->Layers   = ih[maxRateIndex]->Layers;
            ruh->Flags    = 0x00;
            ruh->Position = translate32(fposition);
            ResourceUtilizationEntry* entry = &ruh->Entry[0];
            for(cardinal i = 0;i < newEntries;i++) {
               entry->Utilization  = translate32((cardinal)rint(rupList[i].Utilization * pow(2.0,30.0)));
               entry->FrameRate    = translateToBinary(rupList[i].FrameRate);
               entry->Bandwidth[0] = translate32(rupList[i].Bandwidth);
               for(cardinal j = 1;j < (cardinal)(ih[maxRateIndex]->Layers + 1);j++) {
                  entry->Bandwidth[j] = translate32(rupList[i].LayerBandwidthInfo[j - 1].BytesPerSecond);
               }
               entry = (ResourceUtilizationEntry*)((long)entry +
                                                   (long)(sizeof(ResourceUtilizationEntry) +
                                                          ((ih[maxRateIndex]->Layers + 1) * sizeof(card32))));
            }
            if(write(InputFile,(void*)ruh,sizeof(ruListBuffer)) != (ssize_t)sizeof(ruListBuffer)) {
               cerr << "ERROR: update() - Write error!" << endl;
               close();
               return(false);
            }
            filePosition += sizeof(ruListBuffer);
         }


         // ====== Print progress ===========================================
         char str[256];
         snprintf((char*)&str,sizeof(str),"\x0dWorking... [%1.1f %%]",
            (double)(fposition * 100) / (double)(frames[maxRateIndex] - 1));
         cout << str;
         cout.flush();
      }
      if(indexPosition > 0) {
         index[indexPosition - 1].Length =
            translate32(frames[maxRateIndex] - translate32(index[indexPosition - 1].Position));
      }
      cout << endl;


      // ====== Write index =================================================
      ruIndex = filePosition;
      ResourceUtilizationListIndexHeader rulih;
      rulih.Entries = translate32(indexPosition);
      if(write(InputFile,(void*)&rulih,sizeof(rulih)) != (ssize_t)sizeof(rulih)) {
         cerr << "RROR: update() - Write error!" << endl;
         close();
         return(false);
      }
      const cardinal indexSize = indexPosition * sizeof(ResourceUtilizationListIndexEntry);
      if(write(InputFile,(void*)&index,indexSize) != (ssize_t)indexSize) {
         cerr << "RROR: update() - Write error!" << endl;
         close();
         return(false);
      }
   }


   // ====== Update information strings of the TDTF prefix ==================
   if(config.Title[0] != 0x00) {
      strncpy((char*)&tp->Title,(char*)&config.Title,TDTFPrefix::MaxTitleLength);
      for(cardinal i = strlen(tp->Title);i < TDTFPrefix::MaxTitleLength;i++) {
         tp->Title[i] = 0x00;
      }
   }
   if(config.Copyright[0] != 0x00) {
      strncpy((char*)&tp->Copyright,(char*)&config.Copyright,TDTFPrefix::MaxCopyrightLength);
      for(cardinal i = strlen(tp->Copyright);i < TDTFPrefix::MaxCopyrightLength;i++) {
         tp->Copyright[i] = 0x00;
      }
   }
   if(config.Comment[0] != 0x00) {
      strncpy((char*)&tp->Comment,(char*)&config.Comment,TDTFPrefix::MaxCommentLength);
      for(cardinal i = strlen(tp->Comment);i < TDTFPrefix::MaxCommentLength;i++) {
         tp->Comment[i] = 0x00;
      }
   }
   if(config.URL[0] != 0x00) {
      strncpy((char*)&tp->URL,(char*)&config.URL,TDTFPrefix::MaxURLLength);
      for(cardinal i = strlen(tp->URL);i < TDTFPrefix::MaxURLLength;i++) {
         tp->URL[i] = 0x00;
      }
   }


   // ====== Set utilization time stamps ====================================
   ts.ResourceUtilizationIndex = translate32(ruIndex);
   if(!truncateOnly) {
      tp->UtilizationTimeStamp = translate64(startTS);
      ts.UtilizationTimeStamp  = translate64(getMicroTime());
   }
   if(write(InputFile,(void*)&ts,sizeof(ts)) != (ssize_t)sizeof(ts)) {
      cerr << "ERROR: update() - Write error!" << endl;
      close();
      return(false);
   }
   close();
   return(true);
}


// ###### Copy file #########################################################
bool copy(const char* input, const char* output)
{
   TDTFReader reader;
   if(reader.open(input)) {
      FILE* of = fopen(output,"w");
      if(of != NULL) {
         if(fwrite(reader.InputMemory,reader.InputLength,1,of) == 1) {
            fclose(of);
            return(true);
         }
         fclose(of);
         remove(output);
      }
      else {
         cerr << "ERROR: Unable to create file " << output << "!" << endl;
      }
   }
   else {
      cerr << "ERROR: Unable to open file " << input << "!" << endl;
   }
   return(false);
}


// ###### Main program ######################################################
int main(int argc, char* argv[])
{
   // ===== Check arguments =================================================
   if(argc < 5) {
      cerr << "Usage: " << argv[0] << endl
           << "[Configuration file] [Input Trace File] [Output Trace File] [MPEG|MP3] {-truncateOnly}" << endl;
      exit(0);
   }
   cardinal type         = 0;
   bool     truncateOnly = false;
   if(!(strcasecmp(argv[4],"MPEG")))     type = 1;
   else if(!(strcasecmp(argv[4],"MP3"))) type = 2;
   else {
      cerr << "Wrong type: " << argv[4] << endl;
   }
   for(cardinal i = 5;i < (cardinal)argc;i++) {
      if(!(strcasecmp(argv[i],"-truncateOnly")))  truncateOnly = true;
      else if(argv[i][0] == '-') {
         cerr << "Wrong parameter: " << argv[i] << endl;
         return(false);
      }
   }


   // ====== Open input files, load configuration and trace =================
   if(TraceConfig.load(argv[1],true) == false) {
      cerr << "ERROR: Unable to load configuration!" << endl;
      exit(1);
   }
   if(copy(argv[2],argv[3]) == false) {
      cerr << "ERROR: Unable to copy trace!" << endl;
   }
   TDTFUtilizationUpdater update;
   if(update.open(argv[3],true)) {

      // ====== Initialize MPEGWriterQoSDescription =========================
      MPEGWriterQoSDescription mpegQoSDescription(&TraceConfig);
      MP3WriterQoSDescription  mp3QoSDescription(&TraceConfig);
      TraceQoSDescription* qosDescription = NULL;
      if(type == 1) {
         qosDescription = &mpegQoSDescription;
      }
      else if(type == 2) {
         qosDescription = &mp3QoSDescription;
      }
      else {
         exit(1);
      }
      qosDescription->initTraceDescription(
         &update, update.getPosition(),
         update.getMaxPosition(), update.getMaxFrameRate());


      // ====== Print information ===========================================
      const TDTFPrefix* tp = update.getTDTFPrefix();
      cout << "TDTF Utilizer  - Copyright (C) 2000 Thomas Dreibholz" << endl
           << "----------------------------------------------------" << endl
           << endl;
      cout << "Input File:" << endl;
      cout << "   Title     = " << tp->Title     << endl;
      cout << "   Copyright = " << tp->Copyright << endl;
      cout << "   Comment   = " << tp->Comment   << endl;
      cout << "   URL       = " << tp->URL       << endl;
      if(!truncateOnly) {
         cout << endl;
         TraceConfig.print(cout,true);
      }
      else {
         cout << "Truncate only!" << endl;
      }
      cout << endl;


      // ====== Update all intervals ========================================
      if(update.doUpdate(qosDescription,TraceConfig,truncateOnly) == true) {
         cout << endl << "Congratulations! Your trace file's utilization is complete!" << endl;
      }
      else {
         remove(argv[3]);
      }
   }
   else {
      remove(argv[3]);
      cerr << "ERROR: Unable to load TDTF trace file <" << argv[2] << ">!" << endl;
      exit(1);
   }

   return(0);
}
